import torch
import numpy as np

class Pruner:
    def __init__(self, masked_parameters):
        self.masked_parameters = list(masked_parameters)
        self.scores = {}

    def score(self, model, loss, dataloader, device):
        raise NotImplementedError

    def _global_mask(self, sparsity):
        r"""Updates masks of model with scores by sparsity level globally.
        """
        # # Set score for masked parameters to -inf 
        # for mask, param in self.masked_parameters:
        #     score = self.scores[id(param)]
        #     score[mask == 0.0] = -np.inf

        # Threshold scores
        global_scores = torch.cat([torch.flatten(v) for v in self.scores.values()])
        k = int((1.0 - sparsity) * global_scores.numel())
        if not k < 1:
            threshold, _ = torch.kthvalue(global_scores, k)
            for mask, param in self.masked_parameters:
                score = self.scores[id(param)]
                zero = torch.tensor([0.]).to(mask.device)
                one = torch.tensor([1.]).to(mask.device)
                ## introduce epsilon noise to avoid errors in grasp and snip code due to large blocks sharing same score
                # eps =  0.00000000000000000001
                # eps_noise = torch.FloatTensor(score.size()).uniform_(-eps, eps)
                # score = score + eps_noise

                mask.copy_(torch.where(score <= threshold, zero, one))


    def _global_mask_flow(self, sparsity):
        r"""Updates masks of model with scores by sparsity level globally.
        """

        # Threshold scores
        # global_scores = torch.cat([torch.flatten(v) for v in self.scores.values()])
        # k = int((1.0 - sparsity) * global_scores.numel())
        # if not k < 1:
        ## parameter idx, for vanilla networks, even idx is weight, uneven is bias
        idx = 0
        for mask, param in self.masked_parameters:
            ## for first parameter block draw ~ sparsity many entries at random
            if (idx == 0):
                score = self.scores[id(param)]
                k = int((1.0 - sparsity) * score.numel())
                threshold, _ = torch.kthvalue(torch.flatten(score), k)
                zero = torch.tensor([0.]).to(mask.device)
                one = torch.tensor([1.]).to(mask.device)
                mask.copy_(torch.where(score <= threshold, zero, one))
                ## store weight mask
                maskBuffer = mask
            ## for following block restrict to active paths
            else :
                if ((idx % 2) == 0):
                    score = self.scores[id(param)]
                    print(score.size())
                    print(maskBuffer.size())
                    print((torch.sum(maskBuffer, 1).clamp(max=1)[None,:]).size())
                    k = int((1.0 - sparsity) * score.numel())
                    score = torch.sum(maskBuffer, 1).clamp(max=1)[None,:] * score
                    threshold, _ = torch.kthvalue(torch.flatten(score), k)
                    zero = torch.tensor([0.]).to(mask.device)
                    one = torch.tensor([1.]).to(mask.device)
                    mask.copy_(torch.where(score <= threshold, zero, one))
                    maskBuffer = mask
                else:
                    score = self.scores[id(param)]
                    print(score.size())
                    print(maskBuffer.size())
                    print(maskBuffer.sum(1).clamp(max=1).size())
                    score = torch.sum(maskBuffer, 1).clamp(max=1) * score
                    k = int((1.0 - sparsity) * score.numel())
                    threshold, _ = torch.kthvalue(torch.flatten(score), k)
                    zero = torch.tensor([0.]).to(mask.device)
                    one = torch.tensor([1.]).to(mask.device)
                    mask.copy_(torch.where(score <= threshold, zero, one))
            idx += 1



        ## Forward pass: remember all k shortest paths (by edge) ending in each node
        

        print("---------------")

    
    def _local_mask(self, sparsity):
        r"""Updates masks of model with scores by sparsity level parameter-wise.
        """
        for mask, param in self.masked_parameters:
            score = self.scores[id(param)]
            k = int((1.0 - sparsity) * score.numel())
            if not k < 1:
                threshold, _ = torch.kthvalue(torch.flatten(score), k)
                zero = torch.tensor([0.]).to(mask.device)
                one = torch.tensor([1.]).to(mask.device)
                mask.copy_(torch.where(score <= threshold, zero, one))

    def get_masks(self):

        res = []
        for mask, param in self.masked_parameters:
            res += [(mask, param)]

        return res



    def mask(self, sparsity, scope):
        r"""Updates masks of model with scores by sparsity according to scope.
        """
        if scope == 'global':
            self._global_mask(sparsity)
        if scope == 'local':
            self._local_mask(sparsity)
        if scope == 'flow':
            self._global_mask_flow(sparsity)

    @torch.no_grad()
    def apply_mask(self):
        r"""Applies mask to prunable parameters.
        """
        for mask, param in self.masked_parameters:
            param.mul_(mask)

    def alpha_mask(self, alpha):
        r"""Set all masks to alpha in model.
        """
        for mask, _ in self.masked_parameters:
            mask.fill_(alpha)

    # Based on https://github.com/facebookresearch/open_lth/blob/master/utils/tensor_utils.py#L43
    def shuffle(self):
        for mask, param in self.masked_parameters:
            shape = mask.shape
            perm = torch.randperm(mask.nelement())
            mask = mask.reshape(-1)[perm].reshape(shape)

    def invert(self):
        for v in self.scores.values():
            v.div_(v**2)

    def stats(self):
        r"""Returns remaining and total number of prunable parameters.
        """
        remaining_params, total_params = 0, 0 
        for mask, _ in self.masked_parameters:
             remaining_params += mask.detach().cpu().numpy().sum()
             total_params += mask.numel()
        return remaining_params, total_params


class Rand(Pruner):
    def __init__(self, masked_parameters, dataset):
        super(Rand, self).__init__(masked_parameters)
        self.dataset = dataset

    def score(self, model, loss, dataloader, device):
        for _, p in self.masked_parameters:
            self.scores[id(p)] = torch.randn_like(p)


class Mag(Pruner):
    def __init__(self, masked_parameters, dataset):
        super(Mag, self).__init__(masked_parameters)
        self.dataset = dataset
    
    def score(self, model, loss, dataloader, device):
        for _, p in self.masked_parameters:
            self.scores[id(p)] = torch.clone(p.data).detach().abs_()


# Based on https://github.com/mi-lad/snip/blob/master/snip.py#L18
class SNIP(Pruner):
    def __init__(self, masked_parameters, dataset):
        super(SNIP, self).__init__(masked_parameters)
        self.dataset = dataset

    def score(self, model, loss, dataloader, device):

        # allow masks to have gradient
        for m, _ in self.masked_parameters:
            m.requires_grad = True

        # compute gradient
        for batch_idx, (data, target) in enumerate(dataloader):
            if (self.dataset in ['relu','helix']):
                data, target = data.to(device), target.to(device=device, dtype=torch.float)
            else:
                data, target = data.to(device), torch.squeeze(target).to(device=device, dtype=torch.long)
            output = model(data)
            loss(output, target).backward()

        # calculate score |g * theta|
        for m, p in self.masked_parameters:
            self.scores[id(p)] = torch.clone(m.grad).detach().abs_()
            p.grad.data.zero_()
            m.grad.data.zero_()
            m.requires_grad = False

        # normalize score
        all_scores = torch.cat([torch.flatten(v) for v in self.scores.values()])
        norm = torch.sum(all_scores)
        ## only do if normalization term is > 0
        if (norm > 0):
            for _, p in self.masked_parameters:
                self.scores[id(p)].div_(norm)


# Based on https://github.com/alecwangcq/GraSP/blob/master/pruner/GraSP.py#L49
class GraSP(Pruner):
    def __init__(self, masked_parameters, dataset):
        super(GraSP, self).__init__(masked_parameters)
        self.dataset = dataset
        self.temp = 200
        self.eps = 1e-10

    def score(self, model, loss, dataloader, device):

        # first gradient vector without computational graph
        stopped_grads = 0
        for batch_idx, (data, target) in enumerate(dataloader):
            if (self.dataset in ['relu','helix']):
                data, target = data.to(device), target.to(device=device, dtype=torch.float)
            else:
                data, target = data.to(device), torch.squeeze(target).to(device=device, dtype=torch.long)
            
            output = model(data) / self.temp
            L = loss(output, target)

            grads = torch.autograd.grad(L, [p for (_, p) in self.masked_parameters], create_graph=False)
            flatten_grads = torch.cat([g.reshape(-1) for g in grads if g is not None])
            stopped_grads += flatten_grads

        # second gradient vector with computational graph
        for batch_idx, (data, target) in enumerate(dataloader):
            if (self.dataset in ['relu','helix']):
                data, target = data.to(device), target.to(device=device, dtype=torch.float)
            else:
                data, target = data.to(device), torch.squeeze(target).to(device=device, dtype=torch.long)
            output = model(data) / self.temp
            L = loss(output, target)

            grads = torch.autograd.grad(L, [p for (_, p) in self.masked_parameters], create_graph=True)
            flatten_grads = torch.cat([g.reshape(-1) for g in grads if g is not None])
            
            gnorm = (stopped_grads * flatten_grads).sum()
            gnorm.backward()
        
        # calculate score Hg * theta (negate to remove top percent)
        for _, p in self.masked_parameters:
            self.scores[id(p)] = torch.clone(p.grad * p.data).detach()
            p.grad.data.zero_()

        # normalize score
        all_scores = torch.cat([torch.flatten(v) for v in self.scores.values()])
        norm = torch.abs(torch.sum(all_scores)) + self.eps
        for _, p in self.masked_parameters:
            self.scores[id(p)].div_(norm)


class SynFlow(Pruner):
    def __init__(self, masked_parameters, dataset):
        super(SynFlow, self).__init__(masked_parameters)
        self.dataset = dataset

    def score(self, model, loss, dataloader, device):
      
        @torch.no_grad()
        def linearize(model):
            # model.double()
            signs = {}
            for name, param in model.state_dict().items():
                signs[name] = torch.sign(param)
                param.abs_()
            return signs

        @torch.no_grad()
        def nonlinearize(model, signs):
            # model.float()
            for name, param in model.state_dict().items():
                param.mul_(signs[name])
        
        signs = linearize(model)

        (data, _) = next(iter(dataloader))
        input_dim = list(data[0,:].shape)
        input = torch.ones([1] + input_dim).to(device)#, dtype=torch.float64).to(device)
        output = model(input)
        torch.sum(output).backward()
        
        for _, p in self.masked_parameters:
            self.scores[id(p)] = torch.clone(p.grad * p).detach().abs_()
            p.grad.data.zero_()

        nonlinearize(model, signs)

